//============================================================================
// Name        : RunLCSA_Revista.cpp
// Author      : Hctor
// Version     :
// Copyright   : Your copyright notice
// Description : Hello World in C++, Ansi-style
//============================================================================
#include "interface.h"

using namespace std;

static void *Index;

// Structure with all globals parameters program
typedef struct {
	uchar *seq;				// original sequence (1 byte for symbol)
	ulong n;				// length (numbers of symbols) of sequence
	ulong cantBytesLCSA;	// bytes of index

	char nameFileSeq[100];	// file sequence name
	char nameFileIndex[100];	// Index file name
	char nameFileResult[100];

	double timeExp1;		// average time of experiment 1 with 'cantRepeat' repetitions
	double timeExp2;		// average time of experiment 2 with 'cantRepeat' repetitions
	double timeExp3;		// average time of experiment 3 with 'cantRepeat' repetitions
	double timeConstruction;// CPU time construction of structure (including the SA)

	bool listingTrace;		// true: print all details for console
	bool runTest;
	bool createIndex;
} ParamProgram;

/* macro to detect and notify errors  */
#define IFERROR(error) {{if (error) { cout << stderr << error_index(error) << endl; exit(1); }}}
int read_file(ParamProgram *param);
void print_usage(char *);
double getTime(void);
double getTimeNano(void);
void runExperiments(ParamProgram *param);	// Run experiments and create result files
uint computeCeilingLogNum(ulong num);

// ***********************************************************************************************************
// ******************************************** MAIN PROGRAM *************************************************
// RunLCSA's parameters (8)...
// argv[1]: text file's name								(param.nameFileSeq)
// argv[2]: file result								(nameFileResult)
int main(int argc, char *argv[]) {
	int error;
	double timeStart, timeEnd;

	ParamProgram param;
	param.listingTrace = false;
	param.runTest = false;
	param.createIndex = false;

	if(argc < 4) print_usage(argv[0]);
	strcpy(param.nameFileIndex, argv[1]);
	strcpy(param.nameFileResult, argv[2]);
	int create = atoi(argv[3]);
	if (create ==1){
		param.createIndex = true;
		if(argc < 5) print_usage(argv[0]);
		strcpy(param.nameFileSeq, argv[4]);
	}else
		strcpy(param.nameFileSeq, "");

	cout << "Run LCSA's Parameters..." << endl;
	cout << "Sequence File: " << param.nameFileSeq << endl;
	cout << "Index File: " << param.nameFileIndex << endl;
	cout << "Result File: " << param.nameFileResult << endl;

	if (param.createIndex){
		cout << "Read index file..." << endl;
		error = read_file(&param);
		IFERROR(error);
		if(param.listingTrace) {
			cout <<"Text:" << endl;
			cout << param.seq << endl;
		}

		cout << "Create index..." << endl;
		timeStart = getTime();	// calculate time of construction of structure (including suffix array)
		char *options = NULL;
		error = build_index(param.seq, param.n, options, &Index);
		IFERROR(error);
		timeEnd = getTime(); 	// time finished of construction of structure
		param.timeConstruction = timeEnd - timeStart;
		cout << endl << "CPU time construction: " << param.timeConstruction * 1000.0 << " Milliseconds" << endl;
		cout << endl << "Index created" << endl;

		if (param.runTest){
			cout << "Test Access..." << endl;
			error = testAccess(Index, param.listingTrace);
			IFERROR(error);
			cout << endl << "Test Access OK !" << endl;
		}

		cout << "Save index..." << endl;
		error = save_index(Index, param.nameFileIndex);
		IFERROR(error);
		cout << endl << "Index saved" << endl;
	}else{
		cout << "Load index..." << endl;
		timeStart = getTime();	// calculate time of construction of structure (including suffix array)
		error = load_index(param.nameFileIndex, &Index);
		IFERROR(error);
		timeEnd = getTime(); 	// time finished of construction of structure
		param.timeConstruction = timeEnd - timeStart;
		cout << endl << "CPU time construction: " << param.timeConstruction * 1000.0 << " Milliseconds" << endl;
		cout << endl << "Index loaded" << endl;
	}

	error = index_size(Index, &(param.cantBytesLCSA));
	IFERROR(error);
	cout << "Index size: " <<  param.cantBytesLCSA << endl;

	error = get_length(Index, &(param.n));
	IFERROR(error);
	cout << "Index length: " <<  param.n << endl;

	// Run experiments and create result files
	runExperiments(&param);

	cout << endl << "PROGRAM FINISHED OK !!" << endl;
	cout << "_______________________________________________________________________________________________________" << endl;
	cout << "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$" << endl << endl;

	return 0;

}

//  Opens and reads a text file
int read_file(ParamProgram *param) {
  uchar *text;
  unsigned long t;
  FILE *infile;

  infile = fopen(param->nameFileSeq, "rb"); // rb is for binary: required by DOS
  if(infile == NULL) return 1;

  // store input file length
  if(fseek(infile,0,SEEK_END) !=0 ) return 1;
  param->n = ftell(infile);

  // alloc memory for text (the overshoot is for suffix sorting)
  text = (uchar *) malloc(param->n*sizeof(*text));
  if(text == NULL) return 1;

  cout << "Reading sequence from original file, input_size = " << param->n << endl;
  // read text in one sweep
  rewind(infile);
  t = fread(text, sizeof(*text), (size_t) param->n, infile);
  if(t!=param->n) return 1;
  param->seq = text;
  fclose(infile);

  cout << "file read" << endl;

  return 0;
}

// compute ceiling to logarithm base 2 for num
uint computeCeilingLogNum(ulong num){
	uint lg = 0;
	float lgB = log(num)/log(2);

	if((lgB - (uint)lgB) > 0)
    	lg = (uint)lgB + 1;
    else
    	lg = (uint)lgB;

	return lg;
}

// Experiment 1 - compute A[i]: it gives the average time of CPU of A[i] from the representation of GV, for 0<=i<=n-1
double runExperimentsOne(ParamProgram *param, uint repetitions){
	double t, avgTime;
	ulong *value;
	uint i, j;
	//int error;

	cout << "_______________________________________________________________________________________________________" << endl;
	cout << "Start Experiment 1,  Extract A[i], for " << repetitions << " repetitions" << endl;

	avgTime = 0.0;
	for (i=0; i<repetitions; i++){
		j = (rand() % param->n);
		t = getTimeNano();
		//error = access(Index, j, &value);
		//IFERROR(error);
		value = locate_extract(Index, j, j);
		avgTime += (getTimeNano() - t);
		//cout << " (" << j << ")" << value[0];
		free(value);
	}
	avgTime = avgTime/(double)repetitions;
	cout << endl << "Average CPU time for item : " << avgTime*1000.0 << " Microseconds" << endl;
	cout << "Experiment 1 has finished" << endl;
	cout << "_______________________________________________________________________________________________________" << endl;

	return avgTime;

}

// Experiment 2 - compute A[i..j]: it gives the average time of CPU
double runExperimentTwo(ParamProgram *param, uint itemsExp2, uint repetitions){
	double t, avgTime = 0.0;
	uint i, j;
	//int error;
	ulong *values;

	cout << "Start Experiment 2. Extract A[i..i+" << itemsExp2 << "] for "<< repetitions << " repetitions" << endl;

	for (i=0; i<repetitions; i++){
		j = (rand() % (param->n-itemsExp2+1));
		t = getTimeNano();
		//error = accessInterval(Index, j, j+itemsExp2-1, &values);
		values = locate_extract(Index, j, j+itemsExp2-1);
		t = getTimeNano() - t;
		avgTime += t/(double)repetitions;
		free(values);
	}
	cout << "Average CPU time for item : " << avgTime*1000.0 << " Microseconds" << endl;
	cout << "Experiment 2 has finished" << endl;
	cout << "_______________________________________________________________________________________________________" << endl;

	return avgTime;
}

// Experiment 3 - compute A[i..i+1], A[i..i+2], ..., A[i..i+%d], it gives the average time of CPU
double runExperimentThree(ParamProgram *param, uint itemsExp3, uint repetitions, double * timesItems){
	double t;
	uint r, i, j;
	//int error;
	ulong *values;

	for (i=0; i<itemsExp3; i++)
		timesItems[i] = 0.0;

	cout << "Start Experiment 3. Extract A[i], A[i..i+1], A[i..i+2], ..., A[i..i+" << itemsExp3-1 << "] for "<< repetitions << " repetitions" << endl;
	for (i=0; i<itemsExp3; i++){
		for (r=0; r<repetitions; r++){
			j = (rand() % (param->n-itemsExp3+1));
			t = getTimeNano();
			//error = accessInterval(Index, j, j+i, &values);
			values = locate_extract(Index, j, j+i);
			t = getTimeNano() - t;
			timesItems[i] = timesItems[i] + t/(double)repetitions;
			free(values);
		}
	}

	cout << "CPU time for each c = {1,2,...," << itemsExp3 << "} in Microseconds" << endl;
	for (i=0; i<itemsExp3; i++)
		cout << " " << timesItems[i]*1000.0;

	cout << endl << "Experiment 3 has finished" << endl;
	cout << "_______________________________________________________________________________________________________" << endl;

	return 0;
}

// Run experiments and create result file
void runExperiments(ParamProgram *param){
	double *timesItems = new double[20];
	double time1, time2, time3;

	uint rep = 500000;
	time1 = runExperimentsOne(param, rep);
	time2 = runExperimentTwo(param, 10, rep);
	if (param->n > 200)
		time3 = runExperimentTwo(param, 100, rep);

	runExperimentThree(param, 20, rep, timesItems);

	cout << "space t1 t2 t3" << endl;
	cout << param->cantBytesLCSA/(float)param->n << " " << time1*1000.0 << " " << time2*1000.0 << " " << time3*1000.0 << endl;
	FILE *fp = fopen(param->nameFileResult, "a+" );
	fprintf(fp, "%s %f %G %G %G\n", param->nameFileIndex,  (float)param->cantBytesLCSA/(float)param->n, time1*1000.0, time2*1000.0, time3*1000.0);
	fclose ( fp );

	delete (double *) timesItems;

}

double getTime (void){
	double usertime, systime;
	struct rusage usage;

	getrusage (RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec +
		(double) usage.ru_utime.tv_usec / 1000000.0;

	systime = (double) usage.ru_stime.tv_sec +
		(double) usage.ru_stime.tv_usec / 1000000.0;

	return (usertime + systime);
}

// return in Nanoseconds
double getTimeNano(void){
	double usertime, systime;
	struct rusage usage;

	getrusage (RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec * 1000.0 +
			(double) usage.ru_utime.tv_usec / 1000.0;

	systime = (double) usage.ru_stime.tv_sec * 1000.0 +
			(double) usage.ru_stime.tv_usec / 1000.0;

	return (usertime + systime);
}

// RunLCSA's parameters (8)...
// argv[1]: text file's name								(param.nameFileSeq)
// argv[2]:
// argv[3]: number of access								(param.cantAccess)
void print_usage(char * progname) {
	// usage example (m will be 5, 10 and 15): ./RunLCSA /home/hector/dataTest/data2 /media/Datos/doctorado/Compresion/project/experiments/ 50000 3 5 5 1000000
	cout << "LCSA's usage: " << endl;
	cout << " " << progname << "<index file> <result file> <create index (0/1)> [source file]" << endl;
	exit(1);
}



/*	uchar *pattern = new uchar[5];
	pattern[0] = 'l';
	pattern[1] = 'a';
	pattern[2] = 'l';
	cout << "Locate: " <<  pattern << endl;
	ulong *loc;
	ulong numocc;
	error =  locate(Index, pattern, 2, &loc, &numocc);
	IFERROR(error);
	for (uint i=0; i<numocc; i++){
		cout << " " << loc[i];
	}
*/
